import time
import logging.config
import inspect
import os, os.path
import sys
import logging


def init_logging(filename="logging.ini", defaults=None, base_dir=None):
    """
    init logging system by file
    :param filename:
    :param base_dir:
    :return:
    """
    if base_dir is None:
        base_dir = get_root_dir()
    logging.config.fileConfig(os.path.join(base_dir, filename), defaults=defaults)


def get_root_dir():
    """
    get root file dir of script
    :return:
    """
    cur_frame = inspect.currentframe()
    if cur_frame is None:
        return os.path.abspath(os.getcwd())
    while True:
        root_dir = os.path.dirname(os.path.abspath(cur_frame.f_code.co_filename))
        cur_frame = cur_frame.f_back
        if cur_frame is None:
            break
    return root_dir


def push_cur_dir_into_search_path(idx=0):
    """
    push current path to sys search path
    :param idx: path insert index
    :return:
    """
    sys.path.insert(idx, os.path.abspath(os.getcwd()))


def current_time_in_millis():
    """
    get current time in millis
    :return:
    """
    return int(time.time() * 1000)


cur_ms = current_time_in_millis


def read_line_from_file(filename, mode='r', allow_blank=False, skip_line=0):
    """
    line by line from file
    :param filename: filename
    :param mode: file open mode
    :param allow_blank: whether emit blank line
    :return:
    """
    with open(filename, mode) as fp:
        for line in fp:
            line = line.strip()
            if len(line) == 0 and not allow_blank:
                continue
            if skip_line > 0:
                skip_line -= 1
                continue
            yield line


line_file = read_line_from_file


def asbool(obj):
    """
    convert obj to bool
    :param obj:
    :return:
    """
    if isinstance(obj, basestring):
        obj = obj.strip().lower()
        if obj in ['true', 'yes', 'on', 'y', 't', '1']:
            return True
        elif obj in ['false', 'no', 'off', 'n', 'f', '0']:
            return False
        else:
            raise ValueError("String is not true/false: %r" % obj)
    return bool(obj)


def get_trace():
    import traceback
    return traceback.format_exc()


class Singleton(object):
    _singleton_instance = None

    def __new__(cls, *args, **kwargs):
        if not cls._singleton_instance:
            cls._singleton_instance = super(Singleton, cls).__new__(
                cls, *args, **kwargs)
        return cls._singleton_instance

    @classmethod
    def instance(cls, *args, **kw):
        if cls._singleton_instance is None:
            cls(*args, **kw)
        return cls._singleton_instance



# coding: utf8


class RowDict(dict):
    """
    object support __getitem__ and __getattr__
    """

    def __getattr__(self, name):
        if name in self:
            return self[name]
        else:
            raise AttributeError("No such attribute: " + name)

    def __setattr__(self, name, value):
        self[name] = value


class MysqlEngineWithOursql(object):
    def __init__(self, engine):
        if isinstance(engine, (basestring, str,)):
            from sqlalchemy import create_engine
            self.engine = create_engine(engine)
        else:
            self.engine = engine

    @staticmethod
    def translate_in_stmt(sql, args):
        """
        translate in stmt to oursql qmark format
        eg:
            ` select * from table where id in %s  ,  ((10, 20, 30,)) `
            will translate to
            ` select * from table where id in (?, ?, ?)   , (10, 20, 30,)

        :param sql:
        :param args:
        :return:
        """
        in_stmt_placeholders = []
        smart_args = []
        for arg in args:
            if isinstance(arg, (list, tuple)):
                in_stmt_placeholders.append('(' + ','.join(['?' for x in arg]) + ')')
                for x in arg:
                    assert not isinstance(x, (list, tuple)), "list cannot be nested"
                    smart_args.append(x)
            else:
                smart_args.append(arg)
        if len(in_stmt_placeholders) == 0:
            return sql, args
        smart_sql = sql % tuple(in_stmt_placeholders)
        print smart_sql
        return smart_sql, smart_args

    def _raw_fetch(self, sql, args):
        return self.engine.execute(*self.translate_in_stmt(sql, args))

    def fetch(self, sql, args):
        res = self._raw_fetch(sql, args)
        if not res:
            return []
        return res.fetchall()

    def fetch_row(self, sql, args):
        res = self._raw_fetch(sql, args)
        if not res:
            return res
        li = []
        keys = res.keys()
        lower_keys = [x.lower() for x in keys]
        for row in res:
            origin = dict(zip(keys, row))
            lower = dict(zip(lower_keys, row))
            origin.update(lower)
            li.append(RowDict(origin))
        return li

    def fetchone(self, sql, args):
        res = self.fetch(sql, args)
        if not res:
            return None
        assert len(res) == 1, "call fetchone, but result size is %s" % len(res)
        return res[0]

    def fetchone_row(self, sql, args):
        res = self.fetch_row(sql, args)
        if not res:
            return None
        assert len(res) == 1, "call fetchone, but result size is %s" % len(res)
        return res[0]

    def __getattr__(self, item):
        # fall back to engine
        return getattr(self.engine, item)


MysqlEngine = MysqlEngineWithOursql



# coding: utf8

__author__ = 'fyz'

import os
import os.path
import sys
import glob
import hashlib
from functools import total_ordering
import argparse
from ...utils import cur_ms
from ...conf.conf_loader import load_configuration, ConfigurationContext

from sqlalchemy import create_engine

SCHEMA_TABLE = """
    CREATE TABLE IF NOT EXISTS `schema_version` (
      `version` int(11) PRIMARY KEY,
      `script` varchar(512) NOT NULL,
      `checksum` varchar(64) DEFAULT NULL,
      `time_created` bigint(20) NOT NULL,
      `execution_time` int(11) NOT NULL,
      `status` int(11) NOT NULL DEFAULT 0
    ) ENGINE=InnoDB DEFAULT CHARSET=utf8mb4 COLLATE=utf8mb4_unicode_ci
"""


def info(msg):
    print "INFO %s" % msg


def error(msg):
    print "\033[91mERROR %s\033[0m" % msg


def prompt(msg, cb, *args, **kw):
    res = raw_input(msg + " [Y/N]: ")
    if res.lower() == 'y':
        cb(*args, **kw)


DEFAULT_CONFIG = {
    "ignore_order": False,
    "pattern": "V*.sql"
}


@total_ordering
class Script(object):
    PREFIX = 'V'
    MAGIC_NUMBER = 100000
    VERSION_SPLITTER = "__"

    def __init__(self, location=None, version=None, checksum=None, filename=None, context=None):
        self.context = context
        self.location = location
        self.content = None
        self.filename = filename
        if self.location is not None:
            assert os.path.isfile(self.location), "script location must be file, %s" % location
            _, self.filename = os.path.split(self.location)
            self.original_version, self.version = self.extract_version()
            self.load_content()
        else:
            assert version is not None
            self.version = version
            self.original_version = self.from_magic_version(self.version)
            self.checksum = checksum

    def extract_version(self):
        version, description = self.filename.split(self.VERSION_SPLITTER, 1)
        assert version.startswith(self.PREFIX), "filename should start with %s" % self.PREFIX
        assert len(description) > 0
        original_version = version[1:]
        return original_version, self.generate_magic_version(original_version)

    @staticmethod
    def generate_magic_version(original_version):
        return int(float(original_version) * Script.MAGIC_NUMBER)

    @staticmethod
    def from_magic_version(version):
        return str(version / float(Script.MAGIC_NUMBER))

    @staticmethod
    def calculate_checksum(content):
        m2 = hashlib.md5()
        m2.update(content)
        return m2.hexdigest()

    def load_content(self):
        with open(self.location, 'r') as fp:
            content = '\n'.join(fp.readlines())
            self.content = content
            self.checksum = self.calculate_checksum(self.content)

    def __eq__(self, other):
        return self.version == other.version and self.checksum == other.checksum

    def __ne__(self, other):
        return not self.__eq__(other)

    def __lt__(self, other):
        return self.version < other.version

    def __str__(self):
        return '[ Script %s %s ]' % (self.version, self.filename)

    __repr__ = __str__


class MigrationContext(object):
    LIST_SQL = "select * from schema_version order by version"
    INSERT_SQL = """insert into schema_version values ('%s', '%s', '%s', %s, 0, 0)"""
    INSTALL_SUCCESS_SQL = '''update schema_version set `status` = 1, execution_time=%s where `version` = %s'''
    LIST_UNFINISHED_SQL = '''select version, script, time_created from schema_version where `status` = 0'''

    def __init__(self, cmd, args, config):
        self.cmd = cmd
        self.args = args
        self.config = config

        self.location = config.get('location')
        self.pattern = config.get('pattern')

        self._init_table()

        self._check_unfinished_script()

    def _init_table(self):
        self.engine = create_engine(self.config.get('url'))
        self.engine.execute(SCHEMA_TABLE)

    def _check_unfinished_script(self):
        scripts = self.engine.execute(self.LIST_UNFINISHED_SQL)
        if scripts and scripts.rowcount > 0:
            error("found unfinished scripts")
            for s in scripts:
                error("%s, %s" % (s[0], s[1],))
            sys.exit(1)


    def execute(self):
        self.cmd = self.cmd.replace('-', '_')
        method = getattr(self, "handle_" + self.cmd, None)
        assert method is not None, "No Command <%s>" % self.cmd
        method()

    ##############################################################
    ###########             CMD SECTION            ###############
    ##############################################################

    def handle_list_config(self):
        assert len(self.args) == 0, "arguments error"
        for k, v in self.config.get_inner_value().iteritems():
            print k, v

    def handle_list_scripts(self):
        assert len(self.args) == 0, "arguments error"
        scripts = self._list_all_scripts()
        for s in scripts:
            print s

    def handle_list_installed_scripts(self):
        assert len(self.args) == 0, "arguments error"
        scripts = self._list_installed_scripts()
        for s in scripts:
            print s

    def handle_list_uninstall_scripts(self):
        all_scripts = self._list_all_scripts()
        installed_scripts = self._list_installed_scripts()
        uninstall_scripts = [x for x in all_scripts if x not in installed_scripts]
        for s in uninstall_scripts:
            print s

    def handle_validate(self):
        all_scripts = self._list_all_scripts()
        installed_scripts = self._list_installed_scripts()
        for s in installed_scripts:
            if s not in all_scripts:
                print 'INVALID', s

    def handle_install(self):
        installed_scripts = self._list_installed_scripts()
        versions = [Script.generate_magic_version(x) for x in self.args]
        i_scripts = self._find_script(installed_scripts, *versions)
        has_error = False
        if len(i_scripts) > 0:
            for k, v in i_scripts:
                if v is not None:
                    has_error = True
                    error("%s has been installed" % v.filename)
        if has_error:
            sys.exit(1)

        all_scripts = self._list_all_scripts()
        u_scripts = self._find_script(all_scripts, *versions)
        for k, v in u_scripts:
            if v is None:
                has_error = True
                error("version %s file not exist" % Script.from_magic_version(k))
        if has_error:
            sys.exit(1)

        for k, v in u_scripts:
            prompt("install %s ?" % v.filename, self._install_script, v)

    def handle_migrate(self):
        info("MIGRATE BEGIN")
        installed_scripts = self._list_installed_scripts()
        all_scripts = self._list_all_scripts()
        uninstall_scripts = [x for x in all_scripts if x not in installed_scripts]
        if not self.config.get("ignore_order"):
            self._check_order(installed_scripts, uninstall_scripts)
        assert len(installed_scripts) + len(uninstall_scripts) == len(all_scripts)

        self._install_script(uninstall_scripts)
        info("MIGRATE END, INSTALL %s SCRIPTS" % len(uninstall_scripts))

    ##############################################################
    ###########             CMD SECTION            ###############
    ##############################################################

    @staticmethod
    def _check_script_version_duplicate(scripts):
        if len(scripts) < 2:
            return
        for i in range(len(scripts) - 1):
            assert scripts[i].version != scripts[i + 1].version, \
                "duplicate script version, %s, %s" % (scripts[i].filename, scripts[i + 1].filename,)

    def _list_all_scripts(self):
        current_dir = os.getcwd()
        abs_location = os.path.abspath(os.path.join(current_dir, self.location))
        assert os.path.isdir(abs_location), "location should be a directory"

        os.chdir(abs_location)
        script_files = glob.glob(self.pattern)
        os.chdir(current_dir)

        if len(script_files) == 0:
            return []
        scripts = [Script(location=os.path.join(abs_location, x), context=self) for x in script_files]
        scripts.sort()
        self._check_script_version_duplicate(scripts)
        return scripts

    def _list_installed_scripts(self):
        res = self.engine.execute(self.LIST_SQL)
        if not res:
            return []
        installed_scripts = []
        for row in res:
            tmp = dict(zip(res.keys(), row))
            installed_scripts.append(
                Script(version=tmp['version'], checksum=tmp['checksum'], filename=tmp['script'], context=self))
        installed_scripts.sort()
        return installed_scripts

    def _execute_script(self, script):
        info("install script BEGIN: %s" % script)
        begin_time = cur_ms()

        begin_sql = self.INSERT_SQL % (
            script.version, script.filename, script.checksum, cur_ms())
        self.engine.execute(begin_sql)

        content = script.content.strip().strip(";")
        contents = content.split(";")
        for c in contents:
            c = c.strip()
            if len(c) > 0:
                self.engine.execute(c)
        finish_sql = self.INSTALL_SUCCESS_SQL % (cur_ms() - begin_time, script.version,)
        self.engine.execute(finish_sql)

    @staticmethod
    def _check_order(s1, s2):
        if len(s1) == 0 or len(s2) == 0:
            return

        assert s1[-1].version < s2[0].version, "version not in order, %s not install, but %s has been installed" % (
            s2[0].filename, s1[-1].filename)

    def _install_script(self, script):
        if isinstance(script, list):
            for single_script in script:
                self._install_script(single_script)
        else:
            self._execute_script(script)

    @staticmethod
    def _find_script(scripts, *versions):
        if len(versions) == 0:
            return []
        result = []
        for version in versions:
            found = False
            for script in scripts:
                if script.version == version:
                    result.append((version, script,))
                    found = True
                    break
            if not found:
                result.append((version, None,))
        return result


def main():
    parser = argparse.ArgumentParser(
        description="Python Version Of Flyway. Dependencies: sqlalchemy, oursql/mysql-connector")
    parser.add_argument('cmd', nargs='+', help="command")
    parser.add_argument('--url', help="database url \n \tformat: mysql+oursql(mysqlconnector)://user:pw@host/dbname")
    parser.add_argument('--location', help="script files location")
    parser.add_argument('--pattern', help="script file pattern")
    parser.add_argument('--ignore_order', action="store_true", help="whether to ignore script order")
    parser.add_argument('--config', '-c', help="config file location")

    cmd_conf, _ = parser.parse_known_args()
    cmd = cmd_conf.cmd

    if cmd_conf.config is not None:
        conf = load_configuration(filename=cmd_conf.config)
    else:
        conf = ConfigurationContext()
    for k, v in cmd_conf.__dict__.iteritems():
        if v is not None and k != 'cmd':
            conf.fill_kv(k, v)

    conf.fill_dict(DEFAULT_CONFIG, only_absent=True).lock()

    MigrationContext(cmd[0], cmd[1:], conf).execute()



import os, os.path
import tornado.ioloop
import tornado.web

from ...utils import get_root_dir


class BaseApplication(tornado.web.Application):
    def __init__(self, handlers, **kw):
        base_dir = get_root_dir()
        settings = dict(template_path=os.path.join(base_dir, "./template"),
                        static_path=os.path.join(base_dir, "./static"),
                        debug=False,
                        autoescape=None
                        )
        settings.update(kw)

        super(BaseApplication, self).__init__(handlers, **settings)

    def boot(self, port):
        self.listen(port)
        print 'application start at port %s' % port
        tornado.ioloop.IOLoop.instance().start()


class BaseAuthHandler(tornado.web.RequestHandler):
    def prepare(self):
        _auth = getattr(self, "_auth", None)
        if _auth is None:
            return
        auth = self.get_argument('auth', None)
        if auth is None:
            auth = self.request.headers.get("auth", None)

        if auth is None or auth != _auth:
            self.finish("auth fail")



# coding: utf-8

import os, os.path
import re
import logging
from ..utils import asbool


class TagHandler(object):
    """
    configuration tag handler
    """

    def __init__(self, ctx=None):
        self.ctx = ctx

    def handle(self, line, ctx):
        raise NotImplementedError()


class IncludeTagHandler(TagHandler):
    def handle(self, line, ctx=None):
        if ctx is None:
            ctx = self.ctx
        assert ctx is not None
        filename = line.split()[1].strip()

        ctx.fill_file(os.path.join(os.path.dirname(ctx.current_file), filename), root=False)


class ConfigurationContext(object):
    """
    properties configuration loader

    support include
    """

    def __init__(self, strict=False):
        self._locked = False
        self.strict = strict
        self._inner_value = {}
        self.current_file = None

        self._tag_handlers = {
            "include": IncludeTagHandler()
        }

        self._error_count = 0

    def _check_lock(self):
        if self._locked:
            raise Exception("configuration is locked, cannot modify any more")

    def fill_line(self, line):
        self._check_lock()
        line = line.strip()
        if len(line) == 0:
            return

        if line.startswith('@'):
            # tag
            self._handle_tag(line)
        elif line.startswith("#"):
            # comment
            pass
        else:
            try:
                key, value = re.split("=|:", line, 1)
                self.fill_kv(key.strip(), value.strip())
            except Exception, e:
                if self.strict:
                    self._error_count += 1
                    logging.error("line: %s\t%s" % (line, e,))

        return self

    def add_tag_handler(self, name, tag_handler):
        self._tag_handlers[name] = tag_handler

    def _handle_tag(self, line):
        if len(line) < 2:
            return
        tag_indicator = line[1:].split()[0].strip()
        self._tag_handlers[tag_indicator].handle(line, ctx=self)

    def fill_file(self, filename, root=True):
        self._check_lock()
        assert os.path.exists(filename) and os.path.isfile(filename), "conf file %s not exist" % filename
        self.current_file = filename
        with open(filename, 'r') as fp:
            for line in fp:
                self.fill_line(line)

        if root and self._error_count > 0:
            raise Exception("encounter %s errors when parse config file" % self._error_count)
        return self

    def fill_kv(self, key, val, only_absent=False):
        self._check_lock()
        if only_absent and key in self._inner_value:
            return self
        self._inner_value[key] = val
        return self

    def fill_dict(self, dic, only_absent=False):
        self._check_lock()
        if not isinstance(dic, dict):
            return self
        for k, v in dic.iteritems():
            self.fill_kv(k, v, only_absent=only_absent)
        return self

    @staticmethod
    def _resolve_key_prefix(key, prefix):
        if prefix is None:
            return key
        return '%s.%s' % (prefix, key,)

    def get(self, key, default=None):
        return self._inner_value.get(key, default)

    def get_int(self, key, default=None):
        return int(self.get(key, default))

    def get_bool(self, key, default=None):
        value = self.get(key, default)
        return asbool(value)

    def mget(self, *keys, **kw):
        prefix = kw.get("prefix", None)
        values = []
        for key in keys:
            values.append(self.get(self._resolve_key_prefix(key, prefix)))
        return tuple(values)

    def lock(self):
        self._locked = True
        return self

    def get_inner_value(self):
        return self._inner_value

    def __str__(self):
        return str(self._inner_value)

    __repr__ = __str__


def load_configuration(filename, base_dir=None):
    assert filename is not None, "filename should not be None"
    assert len(filename) > 0
    if base_dir is not None:
        filename = os.path.join(base_dir, filename)
    filename = os.path.abspath(filename)
    return ConfigurationContext(strict=True).fill_file(filename)



import argparse

parser = argparse.ArgumentParser(description="Python Version Of Flyway. Dependencies: sqlalchemy, oursql")
parser.add_argument('cmd', nargs='+', help="command")
parser.add_argument('--url', help="database url")
parser.add_argument('--location', help="script files location")
parser.add_argument('--pattern', help="script file pattern")
parser.add_argument('--debug', action="store_true", help="whether to turn on debug log")
# parser.add_argument('--list-installed', action="store_true", help="list installed scripts")
# parser.add_argument('--list-uninstalled', action="store_true", help="list uninstalled scripts")
# parser.add_argument('--install', help="script version to install")
#
parser.add_argument('--config', help="config file location")
# parser.add_argument('--list-config', action="store_true", help="only list config")
# parser.add_argument('--validate', action="store_true", help="only check")
cmd_conf, _ = parser.parse_known_args()
print cmd_conf.__dict__



from setuptools import setup, find_packages

setup(
    name='skyb',
    packages=find_packages(exclude=['test', 'ez_setup']),
    license='MIT',
    version='0.1.7',
    description='skypiea base',
    author='woshifyz',
    author_email='monkey.d.pandora@gmail.com',
    url='https://github.com/woshifyz/skyb',
    download_url='https://github.com/woshifyz/skyb/tarball/0.1.7',
    keywords=['skypiea', 'base'],
    classifiers=[],
    entry_points=dict(
        console_scripts=[
            'skyfly = skyb.quick.flyway:main',
        ],
    ),
    include_package_data=True,
)
